﻿//****************************************************************
//
// Gimp XCF Content Pipeline Importer
//
// Written by: Louis Ingenthron
// Last Modified: 28.05.2009
//
//****************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace XCFProcessorLib
{
    /// <summary>
    /// Represents a color using Hue, Saturation, Luminosity, and Alpha values.
    /// </summary>
    public class ColorHSL
    {
        /// <summary>
        /// Gets or sets the hue component value.
        /// </summary>
        public byte H { get; private set; }
        /// <summary>
        /// Gets or sets the saturation component value.
        /// </summary>
        public byte S { get; private set; }
        /// <summary>
        /// Gets or sets the luminosity component value.
        /// </summary>
        public byte L { get; private set; }
        /// <summary>
        /// Gets or sets the alpha component value.
        /// </summary>
        public byte A { get; private set; }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="colorRGB">The RGB Color to convert from</param>
        public ColorHSL(Color colorRGB)
        {
            Vector3 hsv;
            Vector4 color = colorRGB.ToVector4();
            float mn, mx, delta;
            mn = Math.Min(color.X, Math.Min(color.Y, color.Z));
            mx = Math.Max(color.X, Math.Max(color.Y, color.Z));
            hsv.Z = mx;
            delta = mx - mn;
            if (mx > 0)
            {
                hsv.Y = delta / mx;
                if (color.X == mx)
                    hsv.X = (color.Y - color.Z) / delta;
                else if (color.Y == mx)
                    hsv.X = 2 + (color.Z - color.X) / delta;
                else
                    hsv.X = 4 + (color.X - color.Y) / delta;
                hsv.X /= 6.0f;
                if (hsv.X < 0)
                    hsv.X += 1.0f;
            }
            else
            {
                hsv.Y = 0;
                hsv.X = 0;
            }
            hsv = HSVtoHSL(hsv);
            H = (byte)(hsv.X * 255f);
            S = (byte)(hsv.Y * 255f);
            L = (byte)(hsv.Z * 255f);
            A = colorRGB.A;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="colorRGB">The RGB Color to convert from</param>
        /// <param name="a">The Alpha component (0-255)</param>
        public ColorHSL(Color colorRGB, byte a)
        {
            Vector3 hsv;
            Vector4 color = colorRGB.ToVector4();
            float mn, mx, delta;
            mn = Math.Min(color.X, Math.Min(color.Y, color.Z));
            mx = Math.Max(color.X, Math.Min(color.Y, color.Z));
            hsv.Z = mx;
            delta = mx - mn;
            if (mx > 0)
            {
                hsv.Y = delta / mx;
                if (color.X == mx)
                    hsv.X = (color.Y - color.Z) / delta;
                else if (color.Y == mx)
                    hsv.X = 2 + (color.Z - color.X) / delta;
                else
                    hsv.X = 4 + (color.X - color.Y) / delta;
                hsv.X /= 6.0f;
                if (hsv.X < 0)
                    hsv.X += 1.0f;
            }
            else
            {
                hsv.Y = 0;
                hsv.X = 0;
            }
            hsv = HSVtoHSL(hsv);
            H = (byte)(hsv.X * 255f);
            S = (byte)(hsv.Y * 255f);
            L = (byte)(hsv.Z * 255f);
            A = a;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="colorRGB">The RGB Color to convert from</param>
        /// <param name="a">The Alpha component (0.0-1.0)</param>
        public ColorHSL(Color colorRGB, float a)
        {
            Vector3 hsv;
            Vector4 color = colorRGB.ToVector4();
            float mn, mx, delta;
            mn = Math.Min(color.X, Math.Min(color.Y, color.Z));
            mx = Math.Max(color.X, Math.Min(color.Y, color.Z));
            hsv.Z = mx;
            delta = mx - mn;
            if (mx > 0)
            {
                hsv.Y = delta / mx;
                if (color.X == mx)
                    hsv.X = (color.Y - color.Z) / delta;
                else if (color.Y == mx)
                    hsv.X = 2 + (color.Z - color.X) / delta;
                else
                    hsv.X = 4 + (color.X - color.Y) / delta;
                hsv.X /= 6.0f;
                if (hsv.X < 0)
                    hsv.X += 1.0f;
            }
            else
            {
                hsv.Y = 0;
                hsv.X = 0;
            }
            hsv = HSVtoHSL(hsv);
            H = (byte)(hsv.X * 255f);
            S = (byte)(hsv.Y * 255f);
            L = (byte)(hsv.Z * 255f);
            A = (byte)(a * 255f);
        }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="h">The Hue component (0.0-1.0)</param>
        /// <param name="s">The Saturation component (0.0-1.0)</param>
        /// <param name="l">The Luminosity component (0.0-1.0)</param>
        public ColorHSL(float h, float s, float l)
        {
            H = (byte)(h * 255f);
            S = (byte)(s * 255f);
            L = (byte)(l * 255f);
            A = 255;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="h">The Hue component (0.0-1.0)</param>
        /// <param name="s">The Saturation component (0.0-1.0)</param>
        /// <param name="l">The Luminosity component (0.0-1.0)</param>
        /// <param name="a">The Alpha component (0.0-1.0)</param>
        public ColorHSL(float h, float s, float l, float a)
        {
            H = (byte)(h * 255f);
            S = (byte)(s * 255f);
            L = (byte)(l * 255f);
            A = (byte)(a * 255f);
        }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="h">The Hue component (0-255)</param>
        /// <param name="s">The Saturation component (0-255)</param>
        /// <param name="l">The Luminosity component (0-255)</param>
        public ColorHSL(byte h, byte s, byte l)
        {
            H = h;
            S = s;
            L = l;
            A = 255;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="h">The Hue component (0-255)</param>
        /// <param name="s">The Saturation component (0-255)</param>
        /// <param name="l">The Luminosity component (0-255)</param>
        /// <param name="a">The Alpha component (0-255)</param>
        public ColorHSL(byte h, byte s, byte l, byte a)
        {
            H = h;
            S = s;
            L = l;
            A = a;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="col">The base HSL color to apply alpha to</param>
        /// <param name="a">The Alpha component (0-255)</param>
        public ColorHSL(ColorHSL col, byte a)
        {
            H = col.H;
            S = col.S;
            L = col.L;
            A = a;
        }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="col">The base HSL color to apply alpha to</param>
        /// <param name="a">The Alpha component (0.0-1.0)</param>
        public ColorHSL(ColorHSL col, float a)
        {
            H = col.H;
            S = col.S;
            L = col.L;
            A = (byte)(a * 255f);
        }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="vector">A Framework.Vector3 containing the Hue, Saturation, and Luminosity values defining a color.</param>
        public ColorHSL(Vector3 vector) : this(vector.X, vector.Y, vector.Z) { }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="vector">A Framework.Vector3 containing the Hue, Saturation, and Luminosity values defining a color.</param>
        /// <param name="a">The Alpha component (0.0-1.0)</param>
        public ColorHSL(Vector3 vector, float a) : this(vector.X, vector.Y, vector.Z, a) { }

        /// <summary>
        /// Initializes a new instance of ColorHSL.
        /// </summary>
        /// <param name="vector">A Framework.Vector3 containing the Hue, Saturation, Luminosity, and Alpha values defining a color.</param>
        public ColorHSL(Vector4 vector) : this(vector.X, vector.Y, vector.Z, vector.W) { }

        /// <summary>
        /// Returns the current color as a Framework.Vector3.
        /// </summary>
        /// <returns>The current color represented as a Framework.Vector3.</returns>
        public Vector3 ToVector3()
        {
            return new Vector3(H / 255f, S / 255f, L / 255f);
        }

        /// <summary>
        /// Returns the current color as a Framework.Vector4.
        /// </summary>
        /// <returns>The current color represented as a Framework.Vector4.</returns>
        public Vector4 ToVector4()
        {
            return new Vector4(H / 255f, S / 255f, L / 255f, A / 255f);
        }

        /// <summary>
        /// Returns an RGB Framework.Color representation of the ColorHSL
        /// </summary>
        /// <returns>A Framework.Color RGB equivalent color</returns>
        public Color ToRGB()
        {
            Vector3 hsv = HSLtoHSV(ToVector3());
            Vector3 rgb;
            int i;
            float f, p, q, t;
            if (hsv.Y == 0)
            {
                rgb.X = hsv.Z;
                rgb.Y = hsv.Z;
                rgb.Z = hsv.Z;
            }
            else
            {
                hsv.X *= 6;
                i = (int)hsv.X;
                f = hsv.X - i;
                p = hsv.Z * (1 - hsv.Y);
                q = hsv.Z * (1 - hsv.Y * f);
                t = hsv.Z * (1 - hsv.Y * (1 - f));
                if (i == 0)
                {
                    rgb.X = hsv.Z;
                    rgb.Y = t;
                    rgb.Z = p;
                }
                else if (i == 1)
                {
                    rgb.X = q;
                    rgb.Y = hsv.Z;
                    rgb.Z = p;
                }
                else if (i == 2)
                {
                    rgb.X = p;
                    rgb.Y = hsv.Z;
                    rgb.Z = t;
                }
                else if (i == 3)
                {
                    rgb.X = p;
                    rgb.Y = q;
                    rgb.Z = hsv.Z;
                }
                else if (i == 4)
                {
                    rgb.X = t;
                    rgb.Y = p;
                    rgb.Z = hsv.Z;
                }
                else
                {
                    rgb.X = hsv.Z;
                    rgb.Y = p;
                    rgb.Z = q;
                }
            }
            return new Color(new Vector4(rgb, A / 255f));
        }

        private static Vector3 HSVtoHSL(Vector3 hsv)
        {
            Vector3 ret;
            ret.X = hsv.X;
            ret.Z = (2f - hsv.Y) * hsv.Z;
            ret.Y = hsv.Y * hsv.Z;
            ret.Y /= (ret.Z <= 1f) ? ret.Z : 2f - ret.Z;
            ret.Z /= 2f;
            return ret;
        }

        private static Vector3 HSLtoHSV(Vector3 hsl)
        {
            Vector3 ret;
            ret.X = hsl.X;
            hsl *= 2f;
            hsl.Y *= (hsl.Z <= 1f) ? hsl.Z : 2f - hsl.Z;
            ret.Z = (hsl.Z + hsl.Y) / 2f;
            ret.Y = (2f * hsl.Y) / (hsl.Z + hsl.Y);
            return ret;
        }
    }
}
