﻿//****************************************************************
//
// Gimp XCF Content Pipeline Importer
//
// Written by: Louis Ingenthron
// Last Modified: 28.05.2009
//
//****************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;


namespace XCFProcessorLib
{
    /// <summary>
    /// Represents texture content built from an XCF
    /// </summary>
    public class XCFContent : BitmapContent
    {
        /// <summary>
        /// The actual pixel data in 4 byte clumps (BGRA)
        /// The start index of any pixel is: (((y * Width) + x) * 4)
        /// </summary>
        private byte[] pixelData;

        /// <summary>
        /// Generates an XCFContent from flattening a source XCF
        /// </summary>
        /// <param name="source">The source XCF to get pixel data from</param>
        public XCFContent(XCFSourceFile source) : base((int)source.CanvasWidth, (int)source.CanvasHeight)
        {
            // need a random for "Dissolve"
            Random r = new Random();

            // make the pixelData array and clear it to TransparentBlack
            pixelData = new byte[source.CanvasWidth * source.CanvasHeight * 4];
            for (int x = 0; x < source.CanvasWidth; x++)
                for (int y = 0; y < source.CanvasHeight; y++)
                {
                    pixelData[(((y * Width) + x) * 4) + 2] += 0;
                    pixelData[(((y * Width) + x) * 4) + 1] += 0;
                    pixelData[(((y * Width) + x) * 4) + 0] += 0;
                    pixelData[(((y * Width) + x) * 4) + 3] += 0;
                }

            // note that the top layer is in [0]
            for(int l=source.Layers.Count-1;l>=0;l--)
                if(source.Layers[l].Visible)
                    for(int x=0;x<source.CanvasWidth;x++)
                        for(int y=0;y<source.CanvasHeight;y++)
                        {
                            //for each pixel on the canvas:

                            Color pixelColor = source.Layers[l].GetPixel(x, y);

                            // dont bother if the alpha is this low
                            // note that the alpha will be 0 if the layer isn't over this pixel
                            if (pixelColor.A < 1)
                                continue;

                            // setup source and dest info for blending
                            Vector3 src = pixelColor.ToVector3();
                            float srcA = (pixelColor.A/255f) * (source.Layers[l].Alpha/255f);
                            Vector3 dest = new Vector3(pixelData[(((y * Width) + x) * 4) + 2] / 255f, pixelData[(((y * Width) + x) * 4) + 1] / 255f, pixelData[(((y * Width) + x) * 4) + 0] / 255f);
                            float destA = pixelData[(((y * Width) + x) * 4) + 3] / 255f;
                            Vector4 result = new Vector4(dest, destA);
                            Layer.BlendMode m = source.Layers[l].Blending;

                            // bottom layer can only be Normal or Dissolve
                            // if it isn't, default it to Normal
                            if (l == source.Layers.Count - 1)
                            {
                                if (m != Layer.BlendMode.Normal && m != Layer.BlendMode.Dissolve)
                                    m = Layer.BlendMode.Normal;
                            }

                            // do blending based on blendmode m
                            switch (m)
                            {
                                case Layer.BlendMode.Normal:
                                    float invSrcA = 1 - srcA;
                                    result = new Vector4(Blend(destA,dest,srcA,src), 1 - (1 - destA) * (1 - srcA));
                                    break;
                                case Layer.BlendMode.Dissolve:
                                    if (r.NextDouble() < srcA)
                                        result = new Vector4(src, 1.0f);
                                    else
                                        result = new Vector4(dest, destA);
                                    break;
                                default:
                                    
                                    result = new Vector4(Composite(destA, dest, srcA, src, m), destA);
                                    break;
                            } 

                            // apply the result of the blend
                            pixelData[(((y * Width) + x) * 4) + 2] = (byte)(result.X * 255f);
                            pixelData[(((y * Width) + x) * 4) + 1] = (byte)(result.Y * 255f);
                            pixelData[(((y * Width) + x) * 4) + 0] = (byte)(result.Z * 255f);
                            pixelData[(((y * Width) + x) * 4) + 3] = (byte)(result.W * 255f);
                        }
        }

        /// <summary>
        /// Blends two values
        /// </summary>
        /// <param name="a1">Destination Alpha</param>
        /// <param name="x1">Destination Value</param>
        /// <param name="a2">Source Alpha</param>
        /// <param name="x2">Source Value</param>
        /// <returns>The blended Value</returns>
        private static float Blend(float a1, float x1, float a2, float x2)
        {
            float k = a2/(1-(1-a1)*(1-a2));
            return (1 - k) * x1 + k * x2;
        }

        /// <summary>
        /// Blends two values
        /// </summary>
        /// <param name="a1">Destination Alpha</param>
        /// <param name="x1">Destination Value vector</param>
        /// <param name="a2">Source Alpha</param>
        /// <param name="x2">Source Value vector</param>
        /// <returns>The blended Value</returns>
        private static Vector3 Blend(float a1, Vector3 x1, float a2, Vector3 x2)
        {
            float k = a2 / (1 - (1 - a1) * (1 - a2));
            return (1 - k) * x1 + k * x2;
        }

        /// <summary>
        /// Blends two values based on specific criteria
        /// </summary>
        /// <param name="a1">Destination Alpha</param>
        /// <param name="y1">Destination Value vector</param>
        /// <param name="a2">Source Alpha</param>
        /// <param name="y2">Source Value vector</param>
        /// <param name="m">The Blend Mode</param>
        /// <returns></returns>
        private static Vector3 Composite(float a1, Vector3 y1, float a2, Vector3 y2, Layer.BlendMode m)
        {
            // hue, saturation, value, and color need special treatment
            // the others fall into a predictable pattern
            if(m == Layer.BlendMode.Hue)
            {
                if (y2.X == y2.Y && y2.Y == y2.Z)
                {
                    // saturation is 0, hue is irrelevant
                    return y1;
                }
                else
                {
                    byte H = (new ColorHSV(new Color(y2.X, y2.Y, y2.Z))).H;
                    byte S = (new ColorHSV(new Color(y1.X, y1.Y, y1.Z))).S;
                    byte V = (new ColorHSV(new Color(y1.X, y1.Y, y1.Z))).V;
                    Color rgb = (new ColorHSV(H, S, V)).ToRGB();
                    Vector3 rgbV = rgb.ToVector3();
                    return new Vector3(Blend(a1, y1.X,Math.Min(a1, a2), rgbV.X), Blend(a1, y1.Y,Math.Min(a1, a2), rgbV.Y), Blend(a1, y1.Z,Math.Min(a1, a2), rgbV.Z));
                }
            }
            else if(m == Layer.BlendMode.Saturation)
            {
                byte H = (new ColorHSV(new Color(y1.X, y1.Y, y1.Z))).H;
                byte S = (new ColorHSV(new Color(y2.X, y2.Y, y2.Z))).S;
                byte V = (new ColorHSV(new Color(y1.X, y1.Y, y1.Z))).V;
                Color rgb = (new ColorHSV(H, S, V)).ToRGB();
                Vector3 rgbV = rgb.ToVector3();
                return new Vector3(Blend(a1, y1.X,Math.Min(a1, a2), rgbV.X), Blend(a1, y1.Y,Math.Min(a1, a2), rgbV.Y), Blend(a1, y1.Z,Math.Min(a1, a2), rgbV.Z));
            }
            else if(m == Layer.BlendMode.Value)
            {
                byte H = (new ColorHSV(new Color(y1.X, y1.Y, y1.Z))).H;
                byte S = (new ColorHSV(new Color(y1.X, y1.Y, y1.Z))).S;
                byte V = (new ColorHSV(new Color(y2.X, y2.Y, y2.Z))).V;
                Color rgb = (new ColorHSV(H, S, V)).ToRGB();
                Vector3 rgbV = rgb.ToVector3();
                return new Vector3(Blend(a1, y1.X,Math.Min(a1, a2), rgbV.X), Blend(a1, y1.Y,Math.Min(a1, a2), rgbV.Y), Blend(a1, y1.Z,Math.Min(a1, a2), rgbV.Z));
            }
            else if (m == Layer.BlendMode.Color)
            {
                byte H = (new ColorHSL(new Color(y2.X, y2.Y, y2.Z))).H;
                byte S = (new ColorHSL(new Color(y2.X, y2.Y, y2.Z))).S;
                byte L = (new ColorHSL(new Color(y1.X, y1.Y, y1.Z))).L;
                Color rgb = (new ColorHSL(H, S, L)).ToRGB();
                Vector3 rgbV = rgb.ToVector3();
                return new Vector3(Blend(a1, y1.X, Math.Min(a1, a2), rgbV.X), Blend(a1, y1.Y, Math.Min(a1, a2), rgbV.Y), Blend(a1, y1.Z, Math.Min(a1, a2), rgbV.Z));
            }
            else
                return new Vector3(Composite(a1, y1.X, a2, y2.X, m), Composite(a1, y1.Y, a2, y2.Y, m), Composite(a1, y1.Z, a2, y2.Z, m));
        }

        /// <summary>
        /// Blends two values based on specific criteria
        /// </summary>
        /// <param name="a1">Destination Alpha</param>
        /// <param name="x1">Destination Value</param>
        /// <param name="a2">Source Alpha</param>
        /// <param name="x2">Source Value</param>
        /// <param name="m">The Blend Mode</param>
        /// <returns></returns>
        private static float Composite(float a1, float x1, float a2, float x2, Layer.BlendMode m)
        {
            return Blend(a1, x1,Math.Min(a1, a2), CompositeHelper(x1, x2, m));
        }

        /// <summary>
        /// Does a simple per-element calculation based on BlendMode
        /// </summary>
        /// <param name="x1">Destination Value</param>
        /// <param name="x2">Source Value</param>
        /// <param name="m">The Blend Mode</param>
        /// <returns></returns>
        private static float CompositeHelper(float x1, float x2, Layer.BlendMode m)
        {
            switch (m)
            {
                case Layer.BlendMode.Multiply:
                    return x1 * x2;
                case Layer.BlendMode.Screen:
                    return 1-(1-x1)*(1-x2);
                case Layer.BlendMode.Overlay:
                    return (1 - x2) * (x1*x1) + x2 * (1 - ((1 - x2)*(1 - x2)) );
                case Layer.BlendMode.Difference:
                    return x1 > x2 ? x1-x2 : x2-x1;
                case Layer.BlendMode.Addition:
                    return Clamp(x1+x2);
                case Layer.BlendMode.Subtract:
                    return Clamp(x1-x2);
                case Layer.BlendMode.DarkenOnly:
                    return Math.Min(x1,x2);
                case Layer.BlendMode.LightenOnly:
                    return Math.Max(x1,x2);
                case Layer.BlendMode.Divide:
                    return Clamp(x1/x2);
                case Layer.BlendMode.Dodge:
                    return Clamp(x1/(1-x2));
                case Layer.BlendMode.Burn:
                    return Clamp(1-(1-x1)/x2);
                case Layer.BlendMode.HardLight:
                    return  (x2 < 0.5) ? (2*x1*x2) : (1-2*((1-x1)*(1-x2)));
                case Layer.BlendMode.SoftLight:
                    return (1-x2)*(x1*x1) + x2*(1-((1-x2)*(1-x2)));
                case Layer.BlendMode.GrainExtract:
                    return Clamp(x1-x2+0.5f);
                case Layer.BlendMode.GrainMerge:
                    return Clamp(x1+x2-0.5f);
            }
            throw new Exception("Unknown Layer Blend mode");
        }

        /// <summary>
        /// Keeps a value between 0.0 and 1.0
        /// </summary>
        /// <param name="x">The value to clamp</param>
        /// <returns>The clamped value</returns>
        private static float Clamp(float x)
        {
            return Math.Min(1, Math.Max(0, x));
        }

        protected override bool TryCopyTo(BitmapContent destinationBitmap, Rectangle sourceRegion, Rectangle destinationRegion)
        {
            byte[] toData = destinationBitmap.GetPixelData();
            for (int x = destinationRegion.X; x < destinationRegion.Width; x++)
                for (int y = destinationRegion.Y; y < destinationRegion.Height; y++)
                {
                    int xx = sourceRegion.X + (int)(((x - destinationRegion.X) / (float)destinationRegion.Width) * sourceRegion.Width);
                    int yy = sourceRegion.Y + (int)(((y - destinationRegion.Y) / (float)destinationRegion.Height) * sourceRegion.Height);
                    Color col = new Color(pixelData[(((yy * Width) + xx) * 4) + 2], pixelData[(((yy * Width) + xx) * 4) + 1], pixelData[(((yy * Width) + xx) * 4) + 0], pixelData[(((yy * Width) + xx) * 4) + 3]);
                    toData[(((y * destinationRegion.Width) + x) * 4) + 2] = col.R;
                    toData[(((y * destinationRegion.Width) + x) * 4) + 1] = col.G;
                    toData[(((y * destinationRegion.Width) + x) * 4) + 0] = col.B;
                    toData[(((y * destinationRegion.Width) + x) * 4) + 3] = col.A;
                }
            destinationBitmap.SetPixelData(toData);
            return true;
        }

        // as far as I can tell, this is not needed for this processor
        protected override bool TryCopyFrom(BitmapContent sourceBitmap, Rectangle sourceRegion, Rectangle destinationRegion)
        {
            return false;
        }

        public override void SetPixelData(byte[] sourceData)
        {
            pixelData = sourceData;
        }

        public override byte[] GetPixelData()
        {
            return pixelData;
        }

        public override bool TryGetFormat(out SurfaceFormat format)
        {
            format = SurfaceFormat.Color;
            return true;
        }
    }
}
