﻿//****************************************************************
//
// Gimp XCF Content Pipeline Importer
//
// Written by: Louis Ingenthron
// Last Modified: 28.05.2009
//
//****************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline;

namespace XCFProcessorLib
{
    internal enum CompressionType
    {
        None=0,     // no compression, each pixel stored
        RLE=1,      // Run Length Encoded compression
        ZLib=2,     // Reserved, but never used
        Fractal=3,  // Reserved but never used
    }

    internal class Layer
    {
        public enum ColorMode
        {
            RGB = 0, // RGB color without alpha; 3 bytes per pixel
	        RGBA = 1,// RGB color with alpha;    4 bytes per pixel
	        Grey = 2,// Grayscale without alpha; 1 byte per pixel
            GreA = 3,// Grayscale with alpha;    2 bytes per pixel
	        Indx = 4,// Indexed without alpha;   1 byte per pixel
	        IndA = 5,// Indexed with alpha;      2 bytes per pixel
        }
        public enum BlendMode
        {
            Normal = 0,// Normal
            Dissolve = 1,// Dissolve (random dithering to discrete alpha)
            Behind = 2,// (Behind: not selectable in the GIMP UI)
            Multiply = 3,// Multiply
            Screen = 4,// Screen
            Overlay = 5,// Overlay
            Difference = 6,// Difference
            Addition = 7,// Addition
            Subtract = 8,// Subtract
            DarkenOnly = 9,// Darken Only
            LightenOnly = 10,// Lighten Only
            Hue = 11,// Hue (H of HSV)
            Saturation = 12,// Saturation (S of HSV)
            Color = 13,// Color (H and S of HSL)
            Value = 14,// Value (V of HSV)
            Divide = 15,// Divide
            Dodge = 16,// Dodge
            Burn = 17,// Burn
            HardLight = 18,// Hard Light
            SoftLight = 19,// Soft Light (XCF version >= 2 only)
            GrainExtract = 20,// Grain Extract (XCF version >= 2 only)
            GrainMerge = 21,// Grain Merge (XCF version >= 2 only)
        }

        public int X, Y;
        public uint Width, Height;

        public ColorMode ColMode;

        public int BytesPerPixel
        {
            get
            {
                return GetBPPForColorMode(ColMode);
            }
        }

        public String Name;

        // the hierarchy and mask pointers
        public uint hptr, mptr;

        public byte Alpha;

        //note that this importer does not handle masks atm
        public bool Visible, ApplyMask;

        public BlendMode Blending;

        public Hierarchy Hierarchy;

        public Layer()
        {
            Alpha = 255;
            Visible = true;
            X = 0;
            Y = 0;
            Blending = BlendMode.Normal;
        }

        /// <summary>
        /// Gets the number of Bytes per Pixel in the specified Color Mode
        /// </summary>
        /// <param name="colMode"></param>
        /// <returns></returns>
        internal static int GetBPPForColorMode(ColorMode colMode)
        {
            if (colMode == ColorMode.RGB)
                return 3;
            if (colMode == ColorMode.RGBA)
                return 4;
            if (colMode == ColorMode.Grey)
                return 1;
            if (colMode == ColorMode.GreA)
                return 2;
            if (colMode == ColorMode.Indx)
                return 1;
            if (colMode == ColorMode.IndA)
                return 2;
            return -1;
        }

        /// <summary>
        /// Searches the tiles in the hierarchy's level and gets the specified pixel
        /// </summary>
        /// <param name="x">Canvas-Space X</param>
        /// <param name="y">Canvas-Space Y</param>
        /// <returns>if (x,y) is inside this layer, the color at the specified position. else, TransparentBlack</returns>
        internal Color GetPixel(int x, int y)
        {
            if (x < X)
                return Color.TransparentBlack;
            if (y < Y)
                return Color.TransparentBlack;
            if (x >= X+Width)
                return Color.TransparentBlack;
            if (y >= Y+Height)
                return Color.TransparentBlack;
            int tileX = (x-X) / 64;
            int tileY = (y-Y) / 64;
            int relaX = (x-X) % 64;
            int relaY = (y-Y) % 64;
            return Hierarchy.Level.Tiles[tileX][tileY].ColorData[relaX][relaY];
        }
    }

    /// <summary>
    /// Not yet implemented or used
    /// </summary>
    internal class Channel
    {

    }

    /// <summary>
    /// A useless wrapper class (thanks Gimp!)
    /// </summary>
    internal class Hierarchy
    {
        public uint Width, Height;
        public uint BPP;
        public Level Level;
    }

    /// <summary>
    /// Holds the 64x64 tiles that have pixel data
    /// </summary>
    internal class Level
    {
        public uint Width, Height;
        public Tile[][] Tiles;
    }

    /// <summary>
    /// A section of pixels up to 64x64
    /// </summary>
    internal class Tile
    {
        /// <summary>
        /// The actual data in this, stored as [x][y]
        /// </summary>
        public Color[][] ColorData;

        /// <summary>
        /// Reads non-encoded tile from a stream and makes a Tile out of it
        /// </summary>
        /// <param name="br">The Stream to read from</param>
        /// <param name="colMode">The enclosing Layer's ColorMode</param>
        /// <param name="width">The width of the tile</param>
        /// <param name="height">The height of the tile</param>
        /// <returns>A new tile of (width,height) filled in with juicy data</returns>
        public static Tile Read(BinaryReader br, Layer.ColorMode colMode, int width, int height)
        {
            int bpp = Layer.GetBPPForColorMode(colMode);
            Tile t = new Tile(width, height);
            for (int y = 0; y < height; y++)
                for (int x = 0; x < width; x++)
                {
                    if(colMode == Layer.ColorMode.RGBA)
                        t.ColorData[x][y] = new Color(br.ReadByte(), br.ReadByte(), br.ReadByte(), br.ReadByte());
                    else if(colMode == Layer.ColorMode.RGB)
                        t.ColorData[x][y] = new Color(br.ReadByte(), br.ReadByte(), br.ReadByte(), 255);
                    else if (colMode == Layer.ColorMode.GreA)
                    {
                        byte val = br.ReadByte();
                        t.ColorData[x][y] = new Color(val, val, val, br.ReadByte());
                    }
                    else if (colMode == Layer.ColorMode.Grey)
                    {
                        byte val = br.ReadByte();
                        t.ColorData[x][y] = new Color(val, val, val, 255);
                    }
                }
            return t;
        }

        /// <summary>
        /// Reads Run Length Encoded tile from a stream and makes a Tile out of it
        /// </summary>
        /// <param name="br">The Stream to read from</param>
        /// <param name="colMode">The enclosing Layer's ColorMode</param>
        /// <param name="width">The width of the tile</param>
        /// <param name="height">The height of the tile</param>
        /// <returns>A new tile of (width,height) filled in with juicy data</returns>
        public static Tile ReadRLE(BinaryReader br, Layer.ColorMode colMode, int width, int height)
        {
            int bpp = Layer.GetBPPForColorMode(colMode);
            Tile t = new Tile(width, height);

            // make a temporary array to hold data
            byte[][][] data = new byte[width][][];
            for(int i=0;i<data.Length;i++)
            {
                data[i] = new byte[height][];
                for(int k=0;k<data[i].Length;k++)
                    data[i][k] = new byte[bpp];
            }
            
            // yes... it has a stream for each byte...
            for(int pixByteInd=0;pixByteInd<bpp;pixByteInd++)
            {
                int pixIndex = 0;
                while (pixIndex < width * height)
                {
                    byte n = br.ReadByte();
                    if (n <= 126)
                    {
                        // short stream of single byte
                        byte val = br.ReadByte();
                        for (int i = 0; i <= n; i++)
                        {
                            data[pixIndex % width][pixIndex / width][pixByteInd] = val;
                            pixIndex++;
                        }
                    }
                    else if (n == 127)
                    {
                        // long stream of single byte
                        int num = (((int)br.ReadByte()) * 256) + (int)br.ReadByte();
                        byte val = br.ReadByte();
                        for (int i = 0; i < num; i++)
                        {
                            data[pixIndex % width][pixIndex / width][pixByteInd] = val;
                            pixIndex++;
                        }
                    }
                    else if (n == 128)
                    {
                        // long stream of different bytes
                        int num = (((int)br.ReadByte()) * 256) + (int)br.ReadByte();
                        for (int i = 0; i < num; i++)
                        {
                            data[pixIndex % width][pixIndex / width][pixByteInd] = br.ReadByte();
                            pixIndex++;
                        }
                    }
                    else if (n >= 129 && n <= 255)
                    {
                        // short stream of different bytes
                        int num = 256 - (int)n;
                        for (int i = 0; i < num; i++)
                        {
                            data[pixIndex % width][pixIndex / width][pixByteInd] = br.ReadByte();
                            pixIndex++;
                        }
                    }
                    else
                        throw new Exception();//wtf?!? byte only has so many values!
                }
            }
            // take that gross array and turn it into Colory goodness
            for (int x = 0; x < width; x++)
                for (int y = 0; y < height; y++)
                {
                    if(colMode == Layer.ColorMode.RGBA)
                        t.ColorData[x][y] = new Color(data[x][y][0], data[x][y][1], data[x][y][2], data[x][y][3]);
                    else if (colMode == Layer.ColorMode.RGB)
                        t.ColorData[x][y] = new Color(data[x][y][0], data[x][y][1], data[x][y][2], 255); 
                    else if (colMode == Layer.ColorMode.GreA)
                        t.ColorData[x][y] = new Color(data[x][y][0], data[x][y][0], data[x][y][0], data[x][y][1]);
                    else if (colMode == Layer.ColorMode.Grey)
                        t.ColorData[x][y] = new Color(data[x][y][0], data[x][y][0], data[x][y][0], 255);
                }
            return t;
        }

        /// <summary>
        /// Creates a blank tile.  Use Static Reader functions to create
        /// </summary>
        /// <param name="w">The tile width in pixels</param>
        /// <param name="h">The tile height in pixels</param>
        private Tile(int w, int h)
        {
            ColorData = new Color[w][];
            for (int i = 0; i < w; i++)
                ColorData[i] = new Color[h];
        }
    }

    /// <summary>
    /// Represents all useful data in an XCF file
    /// </summary>
    public class XCFSourceFile
    {
        internal enum ColorMode
        {
            RGB = 0,
            Greyscale = 1,
            Indexed = 2,
        }

        internal uint CanvasWidth, CanvasHeight;
        internal ColorMode ColMode;
        internal CompressionType Compression;

        internal List<Layer> Layers;
        internal List<Channel> Channels;

        /// <summary>
        /// Initializes a new XCFSourceFile
        /// </summary>
        public XCFSourceFile()
        {
            Layers = new List<Layer>();
            Channels = new List<Channel>();
        }
    }
}
